
#include <arduino.h>
#include "ServerLib.h"

#ifndef CODE_BLOCK_GPin_impl
#ifdef ESP32
void GPin::ledcAnalogWrite(uint8_t channel, uint32_t value, uint32_t valueMax) {
    // calculate duty
    
    uint32_t duty = (LEDC_BASE_FREQ / valueMax) * ((value) < (valueMax) ? (value) : (valueMax));
    //todo: remove all Serial.prints in this class
    Serial.println("Analog Writing duty " + String(duty) + " to channel " + String(channel));
    
    // write duty to LEDC
    ledcWrite(channel, duty);
}
// writes the current value of the pin to the GPO
void GPin::analogWriteEsp32() {
    Serial.println("ledcAnalogWrite " + String(mValue) +" to pin " + String(mNumber) + " on channel " + String(mChannel));

    if (mValue == 0) {
        //sometimes ledcAnalogWrite duty = 0 will not have any effect pins D0 - D7 seem to be ok, but others will have no effect
        ledcAnalogWrite(mChannel, 1, 255);
    }
    ledcAnalogWrite(mChannel, mValue, 255);
    //sigmaDeltaWrite(mChannel, mValue);
    
    
    
}
GPin::GPin(const char*strPinName, PINTYPE pinType, int pinNumber, int pinValue, uint8_t pinChannel) {
    mChannel = pinChannel;
    init(strPinName, pinType, pinNumber, pinValue);
}
#else
GPin::GPin(const char*strPinName, PINTYPE pinType, int pinNumber, int pinValue) {
    init(strPinName, pinType, pinNumber, pinValue);
}
#endif

void GPin::init(const char*strPinName, PINTYPE pinType, int pinNumber, int pinValue) {
        mName = NULL;
        setName(strPinName);
        mType = pinType;
        if (mType == PINTYPE_INPUT_ANALOG || mType == PINTYPE_INPUT_DIGITAL){
            pinMode(pinNumber, INPUT);
            mNumber = pinNumber;
            getValue();//to set the member mValue
        } 
#ifdef ESP32
        else if (pinType == PINTYPE_OUTPUT_ANALOG) {
            ledcSetup(mChannel, LEDC_BASE_FREQ, LEDC_TIMER_13_BIT); //setup the LEDC_CHANNEL which is index
            ledcAttachPin(pinNumber, mChannel); //connect the pin and the LEDC_CHANNEL
            /*sigmaDeltaSetup(mChannel, 312500);
            sigmaDeltaAttachPin(pinNumber, mChannel);*/
            set(pinNumber, pinValue);
        }
#endif
        else { //TYPE IS PINTYPE_OUTPUT_DIGITAL or it is (ESP8266 && PINTYPE_OUTPUT_ANALOG)
            pinMode(pinNumber, OUTPUT);
            set(pinNumber, pinValue);
        }
        
}
GPin::~GPin() { 
    this->destroy(); 
};
void GPin::set(int number, int value) {
    mNumber = number;
    setValue(value);
}
void GPin::setName(const char *newPinName) {
    if (newPinName == NULL) 
        return;
    if (mName != NULL) 
        delete[] mName;
    mName = new char[strlen(newPinName) + 1];
    strcpy(mName, newPinName);
}

void GPin::destroy() {
    if (mName != NULL) {
        delete[] mName;
        mName = NULL;  //no need but feels good :)
    }
}

//String GPin::getName() { return this->getName(); }
String GPin::getName() {
    return String(this->mName);
}
void GPin::setValue(int value) { 
    if (mType == PINTYPE_INPUT_ANALOG || mType == PINTYPE_INPUT_DIGITAL) {
        getValue();  //we cannot set value of a input pin, so we will read instead to set the value
        return;
    }
    if (mType == PINTYPE_OUTPUT_DIGITAL) {
        value = (value > 0) ? HIGH : LOW;
        mValue = value;
        digitalWrite(mNumber, mValue);
        return;
    }

    mValue = value; 

#ifdef ESP32 
    analogWriteEsp32();
#else
    analogWrite(mNumber, mValue);
#endif // ESP8266

    
}

//If input type is PINTYPE_INPUT_DIGITAL or PINTYPE_INPUT_ANALOG and readValueFromHardware is true
//then a read will be maid directly to the hardwarepin. otherwise the old member value will be returned.
int GPin::getValue(bool readValueFromHardware) {
    if (readValueFromHardware && !PINTYPE_INPUT_VIRTUAL){
        if (mType == PINTYPE_INPUT_DIGITAL)
            mValue = digitalRead(mNumber);
        else if (mType == PINTYPE_INPUT_ANALOG)
            mValue = analogRead(mNumber);
    }
    return mValue; 
}
int GPin::getNumber() { 
    return mNumber; 
}
int GPin::getType() {
    
    return mType; 
}

/// <summary>
/// Creates a string with a key and a value, which can be used when populating a JSON object.
/// </summary>
/// <param name="name">Name of the key</param>
/// <param name="value">The integer value</param>
/// <returns>
/// A string with a key and a value.  
/// For example: 
/// "args":12
/// </returns>
String GPin::jsonKeyValue(String key, int value) {
    String str = "\"" + key + "\":" + String(value);
    return str;
}
String GPin::toJsonPinout() {
    return jsonKeyValue(getName(), getNumber());
}
String GPin::toJson() {
    //todo: use jsonKeyValue function here
    String str = "{\"pin\":" + String(mNumber) + "," +
        "\"val\":" + String(getValue()) + "," +
        "\"m\":" + String(mType) + "," +
        "\"name\":\"" + getName() + "\"}";
    return str;
}
#endif //#ifndef CODE_BLOCK_GPin_impl

#ifndef CODE_BLOCK_GPins_impl
boolean GPins::setValue(int pinNumber, int newValue) {
    int i = indexOf(pinNumber);
    if (i < 0) return false;
    Serial.println("setting value of " + String(i) + " to " + String(newValue));
    mPins[i]->setValue(newValue);
    return true;
}
// todo: will we need a removePIn function?
#ifdef ESP32 
int GPins::addPin(const char *strPinName, PINTYPE pinType, int pinNumber, int pinValue) {

    if (pinType == PINTYPE_OUTPUT_ANALOG) {
        mChannelCount++; //mChannel is only used when pin is of type PINTYPE_OUTPUT_ANALOG
    }
    mPins[mLength] = new GPin(strPinName, pinType, pinNumber, pinValue, mChannelCount - 1);
    mLength++;
    return mLength - 1;
}
#else
int GPins::addPin(const char *strPinName, PINTYPE pinType, int pinNumber, int pinValue) {
    mPins[mLength] = new GPin(strPinName, pinType, pinNumber, pinValue);
    mLength++;
    return mLength - 1;
}
#endif

//searches for a pin of a specific number and returns it's index in the array
//if nothing is found the function returns -1
int GPins::indexOf(int pinNumber) {
    Serial.print("index of " + String(pinNumber) + " :");
    if (pinNumber < 0) return -1;
    for (int i = 0; i<mLength; i++) {
        if (pinNumber == mPins[i]->getNumber()) {
            Serial.println(i);
            return i;
        }
    }
    Serial.println(-1);
    return -1;
}
//returns true if a pin with the specific number exists in the array
boolean GPins::exists(int pinNumber) {
    return (indexOf(pinNumber) > -1);
}

// returns a pin with a specific number
GPin *GPins::get(int pinNumber) {
    int i = indexOf(pinNumber);
    if (i < 0) return NULL;
    return mPins[i];
}

// returns a pin value with a specific number
int GPins::getValue(int pinNumber) {
    
    GPin *pin = this->get(pinNumber);
    if (pin == NULL) return -1;
    return pin->getValue();
}

//returns all values of the GPins as an JSON array
String GPins::toJson() {
    String str = "[";
    int i;
    for (i = 0; i < mLength; i++) {
        if (i > 0) {
            str = str + ",";
        }
        str = str + mPins[i]->toJson();
    }
    return str + "]";
}
/// <summary>
/// Returns a name and a number of a pin.
/// </summary>
/// <returns>a key-value string of a name and a pin number ready for an json object</returns>
String GPins::JsonPinout() {
    String str = "[";
    int i;
    for (i = 0; i < mLength; i++) {
        if (i > 0) {
            str = str + ",";
        }
        str = str + mPins[i]->toJsonPinout();
    }
    return str + "]";
}
#endif //CODE_BLOCK_GPins_impl

#ifndef CODE_BLOCK_GUrl_impl
//returns "" if nothing is found after given string
String GUrl::getAfter(String str, String afterMe) {
    int index = str.indexOf(afterMe);

    if (index > -1) {
        index += afterMe.length();
        return str.substring(index);
    }
    return "";
}

//return  -9999 if a string is not a number
int GUrl::toNumber(String str) {
    const int ERROR_NUMBER = -9999;
    int iLen = str.length();
    if (iLen < 1) return ERROR_NUMBER;
    for (int i = 0; i < iLen; i++) {
        if (!isDigit(str[i]))
            return ERROR_NUMBER;
    }
    return str.toInt();
}

String GUrl::removeLastSpaceIfExists(String str) {
    Serial.println();
    int lSpace = str.lastIndexOf(" ");
    if (lSpace > 0)
    {
        Serial.println("remove last space: " + String(lSpace));
        str = str.substring(0, lSpace);
    }
    return str;
}

String GUrl::jsonKeyValue(String key, String value) {
    String   str = "\"" + key + "\":" + value;
    return str;
}
String GUrl::jsonKeyValue(String key, int value) {
    String str = "\"" + key + "\":" + String(value);
    return str;
}

String GUrl::jsonObjectType(unsigned int uiType) {
    String str;
    if (uiType < OBJECTTYPE_COUNT)
        str = String(uiType);
    else
        str = "-1";

    return jsonKeyValue("type", str);
}

String GUrl::makeStatusResponceJson(String jsonPins, String jsonWhitelist, String jsonDate) {
    String str = "{" +
        jsonObjectType(OBJECTTYPE_STATUS) + "," +
        jsonKeyValue("pins", jsonPins) + "," +
        jsonKeyValue("whitelist", jsonWhitelist) + "," +
        jsonKeyValue("date", jsonDate) +
        "}";
    return str;
}
String GUrl::makePostLogPinsJson(String deviceId, String jsonPins) {
    String str = "{" +
        jsonObjectType(OBJECTTYPE_LOG_PINS) + "," +
        jsonKeyValue("pins", jsonPins) + "," +
        jsonKeyValue("deviceId", "\""+deviceId+ "\"") +
        "}";
    return str;
}
/// <summary>
/// Formats a http status code
/// </summary>
/// <param name="uiStatusCode">Number of the http status code to format</param>
/// <returns>A string with the http statuscode number and the status text.</returns>
String GUrl::makeHttpStatusCodeString(unsigned int uiStatusCode) {
    String strCode;
    switch (uiStatusCode) {
    case 200:   strCode = String(uiStatusCode) + " OK";
        break;
    case 400:   strCode = String(uiStatusCode) + "  Bad request";
        break;
    case 416:   strCode = String(uiStatusCode) + " Range Not Satisfiable";
        break;
    default:    strCode = "";
        break;
    }
    return strCode;
}

/// <summary>
/// Extracts pin numbers and values from the given string
/// and sets the pin values according to what was extracted.
/// </summary>
/// <param name="unParsedJson">A endline seperated json values on the form { "3":220}</param>
/// <returns>true if successful otherwhise false</returns>
boolean GUrl::extractAndSetPinsAndValues(const char *unParsedJson, GPins *pinnar) {
    boolean ret = false;
    String line = String(unParsedJson);
    String strPin, strValue;
    int pin, val, iCol, iCom;
    Serial.println("extractAndSetPinsAndValues");
    Serial.println("0" + line);
    line = line.substring(1, line.length() - 1);
    Serial.println("1:" + line);
    line += ",";
    Serial.println("2:" + line);
    line.replace("\"", "");
    line.replace("\"", "");
    Serial.println("3:Line:" + line);
    iCol = line.indexOf(':');
    iCom = line.indexOf(',');
    Serial.println("iCol:" + String(iCol) + "iCom:" + String(iCom));
    while (iCol>0 && iCom>2)
    {
        strPin = line.substring(0, iCol);
        strValue = line.substring(iCol + 1, iCom);
        line.remove(0, iCom + 1);
        Serial.print("Pin  :" + strPin);
        Serial.print("    Value:" + strValue);
        pin = toNumber(strPin);
        val = toNumber(strValue);

        if (pin > -1 && val > -1) {
            if (pinnar->setValue(pin, val))
                Serial.print("    new values set");
            ret = true;
        }
        Serial.println();
        iCol = line.indexOf(':');
        iCom = line.indexOf(',');
    }
    return ret;
}

/// <summary>
/// Extracts pin numbers and values from the given string
/// and adds theyr number to the 
/// </summary>
/// <param name="unParsedJson">A endline seperated json values on the form { "3":220}</param>
/// <param name="pinValues">A container to use when storing pinValues</param>
/// <returns>The number of PinValues saved to the Array</returns>
    uint8_t GUrl::extractPinValues(const char *unParsedJson, PinValue pinValuesArray[], uint8_t arrayLength) {
    uint8_t index = 0;
    String line = String(unParsedJson);
    String strPin, strValue;
    int pin, val, iCol, iCom;
    Serial.println("extractPinValues");
    Serial.println("0" + line);
    line = line.substring(1, line.length() - 1);
    Serial.println("1:" + line);
    line += ",";
    Serial.println("2:" + line);
    line.replace("\"", "");
    line.replace("\"", "");
    Serial.println("3:Line:" + line);
    iCol = line.indexOf(':');
    iCom = line.indexOf(',');
    Serial.println("iCol:" + String(iCol) + "iCom:" + String(iCom));
    while (iCol>0 && iCom>2)
    {
        strPin = line.substring(0, iCol);
        strValue = line.substring(iCol + 1, iCom);
        line.remove(0, iCom + 1);
        Serial.print("Pin  :" + strPin);
        Serial.print("    Value:" + strValue);
        pin = toNumber(strPin);
        val = toNumber(strValue);

        if (pin > -1 && val > -1) {
            pinValuesArray[index].pinNumber = pin;
            pinValuesArray[index].pinValue = val;
            index++;
        }
        Serial.println();
        iCol = line.indexOf(':');
        iCom = line.indexOf(',');
    }
    return index;
}

String GUrl::extractAndReturnIPaddress(const char *unParsedJson) {

    String line = String(unParsedJson);
    int iStart, iEnd;
    iStart = line.indexOf("{\"ipaddress\":\"");
    if (iStart < 0) return "";
    iStart = iStart + 14; //length of {"ipaddress":"
    iEnd = line.indexOf("\"}", iStart);

    if (iEnd <= iStart + 6) return "";//shortest valid ip address is 0.0.0.0
    line = line.substring(iStart, iEnd);
    if (line.length() < 7 || line.length() > 15) return ""; // longest ip address 255.255.255.255
    Serial.println("Processed \"" + line + "\" len=" + String(line.length()));
    return line;
}

String GUrl::jsonRoot(unsigned int uiType, String key, String value) {
    if (uiType == OBJECTTYPE_KEYVALUE_STRING)
        value = "\"" + value + "\"";
    String str = "{" +
        jsonObjectType(uiType) +
        "," +
        jsonKeyValue(key, value) +
        "}";

    return str;
}
#endif //#ifndef CODE_BLOCK_GUrl_impl

#ifndef CODE_BLOCK_KeyValue_impl
KeyValue::KeyValue(String key, String value) { set(KEYVALUE_STRING, key, value, 0, 0); }
KeyValue::KeyValue(String key, int value) { set(KEYVALUE_INT, key, "", value, 0); }
KeyValue::KeyValue(String key, double value) { set(KEYVALUE_DOUBLE, key, "", 0, value); }
String KeyValue::getKey() { return mKey; }

void KeyValue::set(unsigned int keyValueType, String key, String strValue, int iValue, double dValue) {
    mKeyValueType = keyValueType;
    mKey = key;
    mstrValue = strValue;
    miValue = iValue;
    mdValue = dValue;
}
String KeyValue::getValueString(boolean addDoubleQuotationIfString) {
    switch (mKeyValueType) {
    case KEYVALUE_STRING:
        if (addDoubleQuotationIfString)
            return "\"" + mstrValue + "\"";
        else
            return mstrValue;
        break;
    case KEYVALUE_INT:
        return String(miValue);
        break;
    case KEYVALUE_DOUBLE:  return String(mdValue, 3);
        break;
    }
    return "";
}

//returns a keyvalue string pair.
//on error "" is returned.
String KeyValue::toJson() {
    String val;
    return "\"" + mKey + "\":" + getValueString(true);
}
#endif //CODE_BLOCK_KeyValue_impl

#ifndef CODE_BLOCK_Json_impl
Json::Json() {}
Json::~Json() { 
    destory(); 
}
String Json::toJson() {

    String str = "{";
    KeyValue *p;
    for (int i = 0; i < size(); i++) {
        p = get(i);
        if (p != NULL)
        {
            if (i>0) str += ",";
            str += p->toJson();
        }
    }
    return str + "}";
}
//deletes all objects and destoys the list;
void Json::destory() {
    KeyValue *p;
    for (int i = 0; i < size(); i++) {
        p = get(i);
        if (p != NULL)
        {
            Serial.print("destoying: " + p->getKey());
            delay(10);
            delete p;
            Serial.println("setting to 0: ");
            set(i, NULL);
        }
    }
    clear();
}
#endif //CODE_BLOCK_Json_impl

#ifndef CODE_BLOCK_Gtime
boolean GTime::setTime(String strTime) {
    String str;
    String num;
    int i;
    Serial.println("strTime: " + strTime);
    i = strTime.indexOf(", "); if (i < 0) return false; else i += 2;
    str = strTime.substring(i);

    //todo: make following into a function always the

    //get the day
    i = str.indexOf(' '); if (i < 1) return false; i += 1;
    num = str.substring(0, i - 1);
    mDay = toNumber(num);
    str = str.substring(i);

    //get the month
    i = str.indexOf(' '); if (i < 1) return false; i += 1;
    num = str.substring(0, i - 1);
    mMonth = strToMonth(num);
    str = str.substring(i);

    //get the year
    i = str.indexOf(' '); if (i < 1) return false; i += 1;
    num = str.substring(0, i - 1);
    mYear = toNumber(num);
    str = str.substring(i);

    //get the hour
    i = str.indexOf(':'); if (i < 1) return false; i += 1;
    num = str.substring(0, i - 1);
    mHours = toNumber(num);
    str = str.substring(i);
    //get the minute
    i = str.indexOf(':'); if (i < 1) return false; i += 1;
    num = str.substring(0, i - 1);
    mMinutes = toNumber(num);
    str = str.substring(i);

    //get the second
    i = str.indexOf(' '); if (i < 1) return false; i += 1;
    num = str.substring(0, i - 1);
    mSeconds = toNumber(num);
    str = str.substring(i);
}

int GTime::toNumber(String str) {
    const int ERROR_NUMBER = -9999;
    int iLen = str.length();
    if (iLen < 1) return ERROR_NUMBER;
    for (int i = 0; i < iLen; i++) {
        if (!isDigit(str[i]))
            return ERROR_NUMBER;
    }
    return str.toInt();
}
int GTime::strToMonth(String month) {
    if (month.length() != 3) return -1;
    month.toLowerCase();
    if (month == "jan") return  1;
    if (month == "feb") return  2;
    if (month == "mar") return  3;
    if (month == "apr") return  4;
    if (month == "may") return  5;
    if (month == "jun") return  6;
    if (month == "jul") return  7;
    if (month == "aug") return  8;
    if (month == "sep") return  9;
    if (month == "oct") return 10;
    if (month == "nov") return 11;
    if (month == "dec") return 12;
    return -1;
}
String GTime::toString() {
    return  String(mMonth) + "/" +
        String(mDay) + "/" +
        String(mYear) + " " +
        String(mHours) + ":" +
        String(mMinutes) + ":" +
        String(mSeconds);
}
String GTime::toStreng() {
    return  String(mDay) + "." +
        String(mMonth) + "." +
        String(mYear) + " " +
        String(mHours) + ":" +
        String(mMinutes) + ":" +
        String(mSeconds);
}
String GTime::toJson() {
    String str = "{\"year\":" + String(mYear) + "," +
        "\"month\":" + String(mMonth) + "," +
        "\"day\":" + String(mDay) + "," +
        "\"hours\":" + String(mHours) + "," +
        "\"minutes\":" + String(mMinutes) + "," +
        "\"seconds\":" + String(mSeconds) + "}";
    return str;
}
#endif //CODE_BLOCK_Gtime

#ifndef CODE_BLOCK_IPAddressList

IPAddressList::~IPAddressList() { 
    destory();
}

bool IPAddressList::add(const char *strIpAddress) {
    IPAddress *p = new IPAddress();
    p->fromString(strIpAddress);
    if (*p == IPAddress(0, 0, 0, 0) || exists(*p)) {
        delete p;
        return false; //we will not allow "0.0.0.0"
    }
    return LinkedList<IPAddress*>::add(p);
}
bool IPAddressList::add(IPAddress ipAddress) {
    if (exists(ipAddress))
        return false;
    return LinkedList<IPAddress*>::add(new IPAddress(ipAddress));
}
bool IPAddressList::exists(String strIpAddress) {
    IPAddress address;
    address.fromString(strIpAddress);
    return exists(address);
}
bool IPAddressList::exists(IPAddress address) { 
    return indexOf(address) > -1;
}

int IPAddressList::indexOf(IPAddress ipAddress) {
    IPAddress *p;
    for (int i = 0; i < size(); i++) {
        p = get(i);
        if (*p == ipAddress)
            return i;
    }
    return -1;
}
bool IPAddressList::remove(const char *strIpAddress) {
    IPAddress p;
    p.fromString(strIpAddress);
    if (p == IPAddress(0, 0, 0, 0)) {
        return false; //we will not allow "0.0.0.0"
    }
    return remove(p);
}
bool IPAddressList::remove(IPAddress ipAddress) {
    int i = indexOf(ipAddress);
    if (i>-1)
    {
        delete get(i);
        set(i, NULL);
    }
    return LinkedList<IPAddress*>::remove(i);
}

bool IPAddressList::isEmpty() {
    IPAddress *p = get(0);
    bool bRet = (p == NULL);
    return bRet;
}

String IPAddressList::toJson() {
    String str = "[";
    IPAddress *p;
    for (int i = 0; i < size(); i++) {
        p = get(i);
        if (i>0)
            str += ",";
        str += "\"" + p->toString() + "\"";
    }
    str += "]";
    return str;
}

void IPAddressList::destory() {
    IPAddress *p;
    for (int i = 0; i < size(); i++) {
        p = get(i);
        if (p != NULL)
        {
            delay(10);
            delete p;
            set(i, NULL);
        }
    }
    clear();
}
#endif //CODE_BLOCK_IPAddressList
