/*
VoffCon is a system for controlling devices and appliances from anywhere.
It consists of two programs.  A “node server” and a “device server”.
Copyright (C) 2016  Gudjon Holm Sigurdsson

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, version 3 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

You can contact the author by sending email to gudjonholm@gmail.com or
by regular post to the address Haseyla 27, 260 Reykjanesbar, Iceland.
*/

/*
    Board: NodeMCU 1.0 (ESP-12E Module)
*/

#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <ESP8266HTTPClient.h>
#include <ESP8266WebServer.h>
#include <ESP8266mDNS.h>


#include "ServerLib.h";
#include "passivestrip.h";


#define INT_FIRSTNODEMCU 123


GTime startTime;
GPins pinnar;
GUrl  urlTool;
///StringList whiteList; //todo: maybe whiteList should be a IPAddress not String
IPAddressList whiteList;

//////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                          //
//             T H E   V A L U E S   T H A T   M U S T   B E   C H A N G E D                //
//                                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////////

const uint32_t STRIP_PIXEL_COUNT = 300;

// After createding a device in voffcon, you need to put it's deviceId here and upload 
// this program to your ESP2866 device.
const char* deviceId = "5b9e1a148965a1151c921b09";
// Name of the wifi (accesspoint)network
// example: "guttisWiFi"
const char* ssid = "YourWifiName";
// Wifi password
// example: "mypasswordToTheWifi"
const char* password = "YourWifiPassword";
// port number which this device will be operating on.
// example: 5100
const int   PORT = 5100;
// example: 6100
const int   voffconServerPort = 6100;

IPAddress   

// ip address which this device will be operating on.
// Example: "192.168.1.158"     
myIp(192,168,1,19),

//the default gateway which this device will be operating on.
// example: "192.168.1.254"
gateway(192,168,1,254),

// the default gatway on this network.
// On windows goto command
// prompt and type "ipconfig"
// example: "255.255.255.0"
subnet(255,255,255,0),

// example: "192.168.1.127"
voffconServerIp(192,168,1,79);
// Additional information

// If the device is NOT connected, it's light is faint.
// If the device is     connected it's light is bright.

// if you want a special address to be whiteListed you can add it here. 
// remove "//" in the next line, if you want to allow requests(commands) from a specific ipaddress.
// #define WHITELIST_ADDRESS "89.17.157.231"

// boolean grantAccessToEverybody:
// Set to true if you want to allow all clients where the first 3 numbers
// in a client IP address are the same same as myIp (this server IP address).

boolean grantAccessToEverybody = true;

//boolean grantAccessToAllClientsOnSameSubnet:
//Set to true if you want to allow all clients where the first 3 numbers
//in a client IP address are the same same as myIp (this server IP address).
//
boolean grantAccessToAllClientsOnSameSubnet = true;

// boolean grantAccessToFirstCaller:
// set to true if you want to allow the first client to call the "/setup" method
// to be automaticly granted access.  that is, client IP address will be whitelisted.
boolean grantAccessToFirstCaller = true;

PassiveStrip strip = PassiveStrip(STRIP_PIXEL_COUNT, D5, NEO_RGB + NEO_KHZ800);

ESP8266WebServer server(PORT);

void handleRoot() {
    sendText(200, "hello from esp8266!");
}

void sendText(int statusCode, const char *strSend) {
    server.send(statusCode, "text/plain", strSend);
}
void sendJson(int statusCode, const char *strJsonString) {
    server.send(200, "application/json", strJsonString);
}

void handleNotFound() {
    String message = "File Not Found\n\n";
    message += "URI: ";
    message += server.uri();
    message += "\nMethod: ";
    switch (server.method())
    {
        case HTTP_GET:   message += "GET";   break;
        case HTTP_POST:  message += "POST";  break;
        case HTTP_DELETE:message += "DELETE";break;
        case HTTP_PUT:   message += "PUT";   break;
        default: message += "UNKNOWN";
    }
    message += "\nArguments: ";
    message += server.args();
    message += "\n";
    for (uint8_t i = 0; i<server.args(); i++) {
        message += " " + server.argName(i) + ": " + server.arg(i) + "\n";
    }
    server.send(404, "text/plain", message);
}

//////////////////////////////////////////////////////////////////////////////
void handleCustom(){
  Serial.println("handlePins client IP: " + server.client().remoteIP().toString());
  int pin;
  int val;
  if (server.method() == HTTP_POST)
  {   Serial.println("HTTP_POST");
    String name;
    
    Serial.println("arg count " + String(server.args()));
    GPins sentPins;
    for (uint8_t i=0; i<server.args(); i++){
      pin = urlTool.toNumber(server.argName(i));
      val = urlTool.toNumber(server.arg(i));
      Serial.print("("+String(pin)+"="+String(val)+") ");
      if (!(pin > -1 && val > -1)){
        Serial.println("Error converting number!!!!");
        String name = server.argName(i);
        String line = server.arg(i);
        if (  name.equals("plain")   && 
              line.indexOf("{\"") == 0 &&
              line.indexOf('}')+1 == line.length()/*We will not handle nested objects*/
            )
        {
              //let's try to parse a json
              String strPin, strValue; 
              Serial.println("let's try to parse a json");
              int iCol=1, iCom;
              Serial.println("0" + line);
              line = line.substring(1, line.length()-1);
              Serial.println("1:" + line);
              line+=",";
              Serial.println("2:" + line);
              line.replace("\"", "");
              line.replace("\"", "");
              Serial.println("3:Line:" + line);
              iCol = line.indexOf(':');
              iCom = line.indexOf(',');
              Serial.println("while before");
              while(iCol>0 && iCom>2)
              {
                strPin = line.substring(0, iCol);
                strValue = line.substring(iCol+1, iCom);
                line.remove(0, iCom+1);
                Serial.println("Pin=" + strPin + ", value=" + strValue);
                pin = urlTool.toNumber(strPin);
                val = urlTool.toNumber(strValue);
               
                if (pin > -1 && val > -1){
                  sentPins.addPin("",PINTYPE_INPUT_VIRTUAL, pin, val);
                } 
                iCol = line.indexOf(':');
                iCom = line.indexOf(',');
              }
              Serial.println("while end");
              
             
        }
        Serial.println();
      } else {
        if (pin > -1 && val > -1){
          sentPins.addPin("", PINTYPE_INPUT_VIRTUAL, pin, val);
        }
      }

    }//for server.args()
    strip.setValues(&sentPins);
    
  }//HTTP_POST
   server.send(200, "application/json", strip.toJson());
   Serial.println(strip.toJson());
}

/////////////////////////////////////////////////////////////////////////////

void handlePins() {
    Serial.println("handlePins client IP: " + server.client().remoteIP().toString());
    //if (!isAuthorized()) return;
    int pin;
    int val;
    if (server.method() == HTTP_POST)
    {
        Serial.println("HTTP_POST");
        String name;

        Serial.println("arg count " + String(server.args()));
        for (uint8_t i = 0; i<server.args(); i++) {
            Serial.println("Set pin " + server.argName(i) + ": " + server.arg(i));
            pin = urlTool.toNumber(server.argName(i));
            val = urlTool.toNumber(server.arg(i));
            if (pin > -1 && val > -1) {
                pinnar.setValue(pin, val);
            }
            else
            {
                Serial.println("Error converting number!!!!");
                String name = server.argName(i);
                String line = server.arg(i);
                if (name.equals("plain") &&
                    line.indexOf("{\"") == 0 &&
                    line.indexOf('}') + 1 == line.length()// We will not handle nested objects
                    )
                {
                    urlTool.extractAndSetPinsAndValues(line.c_str(), &pinnar);
                }
            }
        }
    }
    String itemJson = urlTool.jsonRoot(OBJECTTYPE_PINS_ARRAY, "pins", pinnar.toJson());
    server.send(200, "application/json", itemJson);
}

void handleStatus() {
    Serial.println("handleStatus");
    if (!isAuthorized()) return;
    int pin;
    int val;
    if (!server.method() == HTTP_GET)
    {
        sendText(401, "Method not supported!");
        return;
    }

    String str = urlTool.makeStatusResponceJson(pinnar.toJson(), whiteList.toJson(), startTime.toJson());
    sendJson(200, str.c_str());
    Serial.println("Sending back");
    Serial.println(str);
}

void handleWhitelist() {
    Serial.println("handleWhitelist ");
    if (!isAuthorized()) return;
    if (server.method() == HTTP_GET)
    {
        sendJson(200, whiteList.toJson().c_str());
        return;
    }
    String key;
    String value;
    if (server.method() == HTTP_POST || server.method() == HTTP_DELETE)
    {
        Serial.print("server.method() == HTTP_POST :");
        Serial.print("server.args():"); Serial.println(server.args());
        if (server.args() == 1) 
        {
            int i = 0;
            Serial.println(" " + server.argName(i) + ": " + server.arg(i));
            key = server.argName(i);
            if (key == "plain") {
                value = urlTool.extractAndReturnIPaddress(server.arg(i).c_str());
                Serial.println("ip: \"" + value + "\"");
                
                boolean bExists = whiteList.exists(value);
                if (server.method() == HTTP_POST) {
                    if (!bExists) {
                        if (value.length() > 6 && whiteList.add(value.c_str()))
                            Serial.println("Whitelisting " + value);
                    }
                }
                else if (server.method() == HTTP_DELETE) {
                    if (whiteList.remove(value.c_str()))
                            Serial.println("Removing from whitelist: " + value);
                }
            }
        }
    }//if (server.method() == HTTP_POST || server.method() == HTTP_DELETE)

    sendJson(200, whiteList.toJson().c_str());
}

String getTime() {
    WiFiClient client;
    while (!!!client.connect("google.com", 80)) {
        Serial.println("connection failed, retrying...");
    }

    client.print("HEAD / HTTP/1.1\r\n\r\n");

    while (!!!client.available()) {
        yield();
    }

    while (client.available()) {
        if (client.read() == '\n') {
            if (client.read() == 'D') {
                if (client.read() == 'a') {
                    if (client.read() == 't') {
                        if (client.read() == 'e') {
                            if (client.read() == ':') {
                                client.read();
                                String theDate = client.readStringUntil('\r');
                                client.stop();
                                return theDate;
                            }
                        }
                    }
                }
            }
        }
    }
    return "";
}

void handleStarted() {
    Serial.println("client IP: " + server.client().remoteIP().toString());
    String itemJson = urlTool.jsonRoot(OBJECTTYPE_DATE, "date", startTime.toJson());
    server.send(200, "application/json", itemJson);
}

// show the mappings of the pins
void handlePinout() {
    Serial.println("client IP: " + server.client().remoteIP().toString());
    server.send(200, "application/json", pinnar.JsonPinout());
}

//returns:
//true :  if the calling client is autorized;
// false:  if calling client is not autorized and send the text
// "You are not authorized to use this function" to the client

boolean isAuthorized() {

    //todo: remove next line when you want everybody to be able to access the server
    if (grantAccessToEverybody) return true;

    //todo: remove next line when you want to make this save
    if (grantAccessToAllClientsOnSameSubnet) {

        Serial.println("client[0]" + server.client().remoteIP()[0]);
        Serial.println("server[0]" + WiFi.localIP()[0]);
        if (server.client().remoteIP()[0] == WiFi.localIP()[0] &&
            server.client().remoteIP()[1] == WiFi.localIP()[1] &&
            server.client().remoteIP()[2] == WiFi.localIP()[2])
        {
            Serial.println("First 3 ip digits match");
            return true;
        }
    }
    String strIP = server.client().remoteIP().toString();
    Serial.println("Client IP: " + strIP);
    delay(10);

    if (whiteList.exists(strIP)) {
        Serial.println("whitelisted ip: " + strIP);
        return true;
    }
    sendText(401, "You are not authorized to use this function");
    return false;
}

void handleSetup() {

    if (grantAccessToFirstCaller && whiteList.size() == 0)
        whiteList.add(server.client().remoteIP().toString().c_str());  //first one who calls this method will be whitelisted
    
    if (!isAuthorized()) return;
    
    sendText(200, "Todo: this function, maybe allow add pins later");
}

String reportIn() {
    GUrl lib;
    Serial.println("Reporting in ");
    String ret = "Fri, 1 Jan 1971 00:00:00 GMT";
    String data = "{" +
        lib.jsonKeyValue("id", "\"" + String(deviceId) + "\",") +
        lib.jsonKeyValue("ip", "\"" + WiFi.localIP().toString() + "\",") +
        lib.jsonKeyValue("port", PORT) +
        "}";

    HTTPClient http;
    String host = voffconServerIp.toString() + ":" + String(voffconServerPort);
    String url = "http://" + host + "/devices/reportin";
    http.begin(url);  //Specify destination for HTTP request
    http.addHeader("Content-Type", "application/json");
    http.addHeader("Connection", "close");
    Serial.println("sending");
    Serial.println(data);

    int httpResponseCode = http.POST(data);   //Send the actual POST request

    if (httpResponseCode>0) {

        String response = http.getString();                       //Get the response to the request

        Serial.println(httpResponseCode);   //Print return code
        Serial.println(response);           //Print request answer
        ret = response; //responce should contain date.toUTCString()

    }
    else {

        Serial.print("Error on sending POST: ");
        Serial.println(httpResponseCode);

    }

    http.end();  //Free resources
    return ret;
}

void setup(void) {
    

    Serial.begin(115200);
    Serial.println("Connecting to : " + String(ssid));
    //WiFi.begin(ssid);
   
     
    //WiFi.config(myIp, gateway, subnet); //this line can be skipped.  only use if you want a specific ip address
    
    
    WiFi.mode(WIFI_STA);
    WiFi.begin(ssid, password);
    Serial.println("");
    // Wait for connectionc
    while (WiFi.status() != WL_CONNECTED) {
        delay(500);
        Serial.print(".");
    }
    
    Serial.println("");
    Serial.print("Connected to ");
    Serial.println(ssid);
    Serial.print("IP  address: "); Serial.println(WiFi.localIP());
    Serial.print("Mac address: "); Serial.println(WiFi.macAddress());

    /*if (MDNS.begin("esp8266")) {
        Serial.println("MDNS responder started");
    }*/

    //SETTING_UP_WHITELIST_START
    //Do not remove line, here whitelist ip's will be added by VoffCon Node server
    //SETTING_UP_WHITELIST_END

    startTime.setTime(reportIn());
    
    server.on("/", handleRoot);
    server.on("/pins", handlePins);
    server.on("/whitelist", handleWhitelist);
    server.on("/started", handleStarted);
    server.on("/status", handleStatus);
    server.on("/setup", handleSetup);
    server.on("/pinout", handlePinout);
    server.on("/custom", handleCustom);

    //todo: spurning um hvort við bætum við /addpins eða    /add
    //todo: spurning um hvort við bætum við /removepins eða /remove

    server.on("/inline", []() {
        server.send(200, "text/plain", "this works as well");
    });

    server.onNotFound(handleNotFound);

    server.begin();
    Serial.println("HTTP server started");
    int startState = 700;
	//SETTING_UP_PINS_START
    PINTYPE type = PINTYPE_OUTPUT_ANALOG;
    pinnar.addPin("D0", type, D0, startState);
    pinnar.addPin("D1", type, D1, startState);
    pinnar.addPin("D2", PINTYPE_INPUT_ANALOG, D2, 512);
    pinnar.addPin("D3", PINTYPE_INPUT_DIGITAL, D3, startState);
    pinnar.addPin("D4", PINTYPE_OUTPUT_DIGITAL, D4, startState);
    pinnar.addPin("D5", type, D5, 123);
    pinnar.addPin("D6", type, D6, startState);
    pinnar.addPin("D7", type, D7, startState);
    pinnar.addPin("D8", type, D8, startState);
	//SETTING_UP_PINS_END
    
    Serial.println(pinnar.toJson());
    strip.setProgram(4);
    WiFi.setSleepMode(WIFI_NONE_SLEEP);
}

void loop(void) {
    server.handleClient();
    strip.run();
}
