// File created: 2018-12-04
#include "passivestrip.h"

PassiveStrip::PassiveStrip(int16_t numberOfPixels, uint8_t stripClockPin, neoPixelType neoType) {
    init(numberOfPixels, stripClockPin, neoType);
}
PassiveStrip::~PassiveStrip() {
    delete pixels;
}

int PassiveStrip::randomNumber(int startNum, int endNum) {
  randomSeed(ESP.getCycleCount());
  return random(startNum, endNum);
}

void PassiveStrip::resetColors(){
 //colors[0] = pixels->Color(  0,   0,   0);
	colors[0] = pixels->Color(255,   0,   0);
	colors[1] = pixels->Color(  0, 255,   0);
	colors[2] = pixels->Color(  0, 255, 255);
	colors[3] = pixels->Color(255,   0, 255);
	colors[4] = pixels->Color(255,  60,   0);
	colors[5] = pixels->Color(  0,   0, 255);
  //colors[0] = pixels->Color(255, 255, 255);
}

void PassiveStrip::reset(){
	resetColors();
	tickDelay = 10;
	setProgram(1);
	valueWork1 = valueWork2 = valueWork3 = value1 = value2 = 0;
}

void PassiveStrip::init(int16_t numberOfPixels, uint8_t stripClockPin, neoPixelType neoType) {
    pixels = new Adafruit_NeoPixel(numberOfPixels, stripClockPin, neoType);
	pixels->begin();
	pixelCount = numberOfPixels;
	reset();
	itsTime = millis() + tickDelay;
}

void PassiveStrip::setDelay(unsigned long delay) {
	tickDelay = delay;
	itsTime = millis() + tickDelay;
}

sColor PassiveStrip::decodeColor(uint32_t uiColor) {
   sColor color;
   color.r = (uint8_t)(uiColor >> 16);
   color.g = (uint8_t)(uiColor >>  8);
   color.b = (uint8_t)uiColor;
   return color;
}

 uint32_t PassiveStrip::encodeColor(sColor color) {
   uint32_t uiColor = (uint32_t)color.r << 16 |
                      (uint32_t)color.g <<  8 |
                      (uint32_t)color.b;
   return uiColor;
}

void PassiveStrip::setStripColor(uint32_t color, bool showNow){
	for(int x = 0; x < pixelCount; x++) {
        pixels->setPixelColor(x, color);
	}
	if (showNow)
		pixels->show();
}

void PassiveStrip::pixelsOff(){
	setStripColor(0);
}

uint32_t PassiveStrip::SoftChangePixel(uint32_t uiColorFrom, uint32_t uiColorTo, bool alsoChangeBack, int index, int count){

  if (index == count-1){
    return uiColorFrom;
  }

  sColor fromColor   = decodeColor(uiColorFrom);
  sColor toColor  = decodeColor(uiColorTo);
  int    rDiff = toColor.r - fromColor.r, 
         gDiff = toColor.g - fromColor.g,
         bDiff = toColor.b - fromColor.b;

  float pros = ((index + 1)*(1 + alsoChangeBack)) / (float)count;
  
  if (pros > 1) {
      pros=2 - pros;
  }

  fromColor.r=(fromColor.r +(rDiff * pros));
  fromColor.g=(fromColor.g +(gDiff * pros));
  fromColor.b=(fromColor.b +(bDiff * pros));

  return encodeColor(fromColor);
  
}


void PassiveStrip::setProgram(uint programIndex) {
	if (programIndex < programCount()) {
		selectedProgram = programIndex;
		Serial.println("Program " + String (selectedProgram) + "");
	}
}

void PassiveStrip::programSwipeUp(uint32_t color0, uint32_t color1){
    if (stepIndex > 0)
		pixels->setPixelColor(stepIndex-1, color1);
	else {
		pixels->setPixelColor(pixelCount - 1, color1);
	}
	pixels->setPixelColor(stepIndex, color0);
    pixels->show();
}
void PassiveStrip::programSwipeDown(uint32_t color0, uint32_t color1){
	int x = (pixelCount -1) - stepIndex;
    if ( (  x+1  ) < pixelCount)
		pixels->setPixelColor(x+1, color1);
	else {
		pixels->setPixelColor(0, color1);
	}
	pixels->setPixelColor(x, color0);
    pixels->show();
}
void PassiveStrip::programCandy(){
	uint index = 0;
    for(int x = 0; x < pixelCount; x++) {
		if (index >= colorCount()) 
			index = 0;
        pixels->setPixelColor(x, colors[index]);
		index++;
	}
	pixels->show();
}

void PassiveStrip::programRandomPixel(bool showPixels){
	int numberOfPixels = value1;
	if (numberOfPixels < 1 || numberOfPixels >= pixelCount)
		numberOfPixels = 1;

	setStripColor(colors[0], false);
	int set = pixelCount / numberOfPixels;
	int from, to, num;
	for (int i = 0; i < numberOfPixels; i++) {
		from = i   * set;
		to   = (i+1) * set;
		num = randomNumber(from, to);
		if (num > pixelCount-1)
			num = pixelCount-1;
		pixels->setPixelColor(num, colors[1]);
	}
	
	if (showPixels)
		pixels->show();
}

void PassiveStrip::programFixedColors(uint32_t color0, uint32_t color1){
	
	uint32_t color;
	for(int i = 0; i < pixelCount; i++) {
		if (i >= value1 && i <= value2)
			color=color1;
		else 
			color=color0;
        pixels->setPixelColor(i, color);
	}
	pixels->show();
	
}

#define SOFTSTEPS 60

void PassiveStrip::programRandomSoftPixel(){

	
	int step,count;

	if (pixelCount < SOFTSTEPS) {
		step = stepIndex;
		count = pixelCount;
	} else {
		step = stepIndex % SOFTSTEPS;
		count = SOFTSTEPS;
	}

	if (value2 > 0 && valueWork3 > 0 && valueWork2 % value2 != 0) {
		if (step == 0) {
			valueWork2++;
			setStripColor(colors[0]);
		}
		
		return;
	}

	if (step == 0) {
		programRandomPixel(false);
		valueWork3 = !valueWork3;
		if (valueWork1 == 0 ){
			valueWork2++;
		}
	}

	uint32_t  softColor = SoftChangePixel(colors[0], colors[1], true, step, count);
	for(int i = 0; i < pixelCount; i++) {
		if ( pixels->getPixelColor(i) != colors[0] ) {
			if (softColor != colors[0])
				pixels->setPixelColor(i, softColor);
		}
	};
	pixels->show();
}

void PassiveStrip::run() {
	if (millis() > itsTime )
    {
        itsTime = millis() + tickDelay;
        runProgram(selectedProgram);
        stepIndex++;
        if (stepIndex>=pixelCount){
            stepIndex = 0;
        }
    }
}

void PassiveStrip::setValues(GPins *sentPins){
	int pin, val;
	GPin *pPin;
	 for(int i = 0; i < sentPins->count(); i++){
      pPin = sentPins->getByIndex(i);
      pin = pPin->getNumber();
      val = pPin->getValue();
	  if (pin > 199 && pin < 200+COLORCOUNT) {
		  int cIndex = pin-200;
		  colors[cIndex] = val;
	  } else {
		  switch (pin) {
			  case 101 : tickDelay = val; 
						 break;
			  case 102 : 
			  			 if (val == 5) {
						     reset();
							 return;
						 }
						 else
						     setProgram(val);
			  			break;
			  case 103 : 
			  			 value1 = val;
			  			 break;
			  case 104 : 
			  			 value2 = val;
			  			 break;

			  case 100 : 
			  default  : 
			  			 Serial.println("Todo: what to do if pin: "+ String(pin) +" and  val: " + String(val)); 
		  }
	  }
      //pinnar.setValue(pin, val);
      
    }
}

String PassiveStrip::toJson() {
	String ret = "";
	int cIndex = 0;
	ret+= "{";
	ret+="\"" +String(  cIndex)+ "\":" +String(colors[cIndex]);
	ret+=",\"" +String(++cIndex)+ "\":" +String(colors[cIndex]);
	ret+=",\"" +String(++cIndex)+ "\":" +String(colors[cIndex]);
	ret+=",\"" +String(++cIndex)+ "\":" +String(colors[cIndex]);
	ret+=",\"" +String(++cIndex)+ "\":" +String(colors[cIndex]);
	ret+=",\"" +String(++cIndex)+ "\":" +String(colors[cIndex]);
	ret+=",\"delay\":" +String(tickDelay);
	ret+=",\"com\":"   +String(selectedProgram);
	ret+=",\"value1\":"+String(value1);
	ret+=",\"value2\":"+String(value2);
	ret+= "}";

	return ret;
}

void PassiveStrip::runProgram(uint programIndex){

    switch(programIndex) {
        case 0: pixelsOff();  
				break;
		case 1: setStripColor(colors[0]);  
				break;
		case 2: programSwipeUp(colors[0], colors[1]);  
				break;
		case 3: programSwipeDown(colors[2], colors[3]);  
				break;
		case 4: programCandy();  
				break;
		case 5: reset();  //this is not called because setValues calls reset
				break;
		case 6: programRandomPixel(true);
				break;
		case 7: programRandomSoftPixel();
				break;
		case 8: programFixedColors(colors[0], colors[1]);
				break;
    }
}

uint PassiveStrip::programCount(){
	return 9;
}
