// File created: 2018-12-04
#ifndef PASSIVESTRIP
#define PASSIVESTRIP

#include <Arduino.h>
#include <Adafruit_NeoPixel.h>

#include "ServerLib.h"

#define COLORCOUNT 6

#define DIMSTEPS 60

struct sColor {
   uint8_t r;
   uint8_t g; 
   uint8_t b;
};

class PassiveStrip {
    private:
		int               stepIndex       = 0;
		int 			  value1    	  = 0;
		int 			  value2    	  = 0;
		int 			  valueWork1   	  = 0;
		int 			  valueWork2   	  = 0;
		int 			  valueWork3   	  = 0;
		int				  pixelCount;
        unsigned long     itsTime;
		unsigned long     tickDelay       = 50;
		uint              selectedProgram = 0;
		uint32_t          colors[COLORCOUNT];
        Adafruit_NeoPixel *pixels;
		
		void init(int16_t numberOfPixels, uint8_t stripPin, neoPixelType neoType);
		void resetColors();
		void reset();
		void runProgram(uint programIndex);


		void programSwipeUp(uint32_t color0, uint32_t color1);
		void programSwipeDown(uint32_t color0, uint32_t color1);
		void programCandy();
		void programRandomPixel();
		void programRandomSoftPixel();
		void programFixedColors(uint32_t color0, uint32_t color1);
	    int	 randomNumber(int startNum, int endNum);
		sColor decodeColor(uint32_t uiColor);
		uint32_t encodeColor(sColor color);
		uint32_t SoftChangePixel(uint32_t uiColorFrom, uint32_t uiColorTo, bool alsoChangeBack, int index, int count);
        
    public:
        
        PassiveStrip(int16_t numberOfPixels, uint8_t stripPin, neoPixelType neoType=NEO_GRB + NEO_KHZ800);
        ~PassiveStrip();
		
		void run();
		void setDelay(unsigned long tickDelay);
		void setStripColor(uint32_t color, bool showNow = true);
		void setProgram(uint programIndex);
	
		uint programCount();
		uint colorCount() {return COLORCOUNT;};
		void pixelsOff();
		void setValues(GPins *sentPins);
		String toJson(); 
		
};



#endif